#ifndef __Node__H
#define __Node__H

#include <windows.h>
#include <commctrl.h>

#pragma warning(disable:4786)
#include <string>
#include <vector>
#include <map>
using namespace std;

#include "MathLib.h"

//==========================================================================================

class CNode {

public: // construction/destruction

	CNode (CNode *pParent) : m_eType(eUndefined),
		m_strName("Unnamed"),
		m_pParent(pParent),
		m_iIndex(-1)
	{
		if ( pParent )
			pParent->m_vChildren.push_back(this);

		m_pSceneRoot = this;
		while ( m_pSceneRoot->m_pParent )
			m_pSceneRoot = m_pSceneRoot->m_pParent;
	}

	virtual ~CNode ()
	{
		for ( int i=0; i<NumberOfChildren(); i++ )
			delete GetChildNode(i);
	}

public: // attributes

	enum ENodeTypes {
		eUndefined,
		eSceneRoot,
		eGroup,
		eGeomObj
	} m_eType;

	string		m_strName;

	CNode		*m_pSceneRoot;
	CNode		*m_pParent; // "Scene Root" ha un NULL parent
	int			m_iIndex; // 0 = "Scene Root", -1 = undefined

public: // operations

	int NumberOfChildren()			{ return m_vChildren.size(); }
	CNode *GetChildNode (int id)	{ return m_vChildren[id]; }

	virtual void EvalWorldState (TimeValue time)
	{
		for ( int i=0; i<NumberOfChildren(); i++ )
			GetChildNode(i)->EvalWorldState(time);
	}

	enum ERenderMode {
		eWireframe,
		eSolid,
		eTexture,
		eSolidPlusWireframe,
		eWireframeSpecialGreen,
		eWireframeSpecialRed
	};

	virtual void Render (ERenderMode eMode)
	{
		for ( int i=0; i<NumberOfChildren(); i++ )
			GetChildNode(i)->Render(eMode);
	}

	virtual BOOL IsChild (CNode *pndNode)
	{
		if ( pndNode == this )
			return TRUE;
		else
			for ( int i=0; i<NumberOfChildren(); i++ )
				if ( GetChildNode(i)->IsChild(pndNode) )
					return TRUE;
		return FALSE;
	}

protected: // implementation

	vector<CNode *>				m_vChildren;

	virtual void PostProcess ()
	{
		for ( int i=0; i<NumberOfChildren(); i++ )
			GetChildNode(i)->PostProcess();
	}

	virtual void DumpInTreeCtrl (HWND hTreeCtrl, HTREEITEM hParent = TVI_ROOT)
	{
		for ( int i=0; i<NumberOfChildren(); i++ )
			GetChildNode(i)->DumpInTreeCtrl(hTreeCtrl, hParent);
	}

protected: // implementation helper(s)

	static HTREEITEM InsertItem (HWND hTreeCtrl, int iImage, LPTSTR pszText, HTREEITEM hParent = TVI_ROOT, LPARAM lParam = 0)
	{
		TVINSERTSTRUCT		tvisIS;

		tvisIS.hParent = hParent;
		tvisIS.hInsertAfter = TVI_LAST;
		tvisIS.item.mask = TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM | TVIF_TEXT;
		tvisIS.item.iImage = iImage;
		tvisIS.item.iSelectedImage = iImage;
		tvisIS.item.lParam = lParam;
		tvisIS.item.pszText = pszText;

		return (HTREEITEM)::SendMessage(hTreeCtrl, TVM_INSERTITEM, 0, (LPARAM)&tvisIS);
	}
};

//==========================================================================================

class CSceneRoot : public CNode {

public: // construction/destruction

	CSceneRoot (anim_stream_desc_t *pAnim);
	virtual ~CSceneRoot () { free(m_pASD); }

public: // attributes

	void SetTreeCtrlWithInfo (HWND hTreeCtrl);

		// SetTreeCtrlWithInfo image list details

			//	0		information
			//	1		materials
			//	2		material
			//	3		scene root
			//	4		group
			//	5		geom obj
			//	6		mesh
			//	7		editable item

public: // operations

	virtual void Render (ERenderMode eMode)
	{
		CNode::Render( eMode == eWireframeSpecialGreen ? eWireframeSpecialRed : eMode );
	}

public: // attributes + state

	//-----------------------------------------
	struct SMaterial {
		int			m_iIndex;
		string		m_strName;
		int			m_iTris;
	};
	//-----------------------------------------

	anim_stream_desc_t	*m_pASD;
	global_info_t		*m_pGI;

	scene_matlist_t		*m_pSML;
	vector<SMaterial>	m_vMaterials;

	map<int, CNode *>	m_mId2Obj;
	BYTE				*m_pAnimPos;

	BOOL				m_bCorrupt;

	int					m_iNumOfNodes;

protected: // implementation

	virtual void DumpInTreeCtrl (HWND hTreeCtrl, HTREEITEM hParent = TVI_ROOT)
	{
		char		szString[1024];
		sprintf(szString, "%s (null)", m_strName.c_str());
		HTREEITEM	hItem = InsertItem(hTreeCtrl, 3, szString, hParent, (LPARAM)(CNode*)this);
		CNode::DumpInTreeCtrl(hTreeCtrl, hItem);
	}
};

//==========================================================================================

class CGroup : public CNode {

public: // construction/destruction

	CGroup (CNode *pParent, scene_group_t *pGroupHead);
	virtual ~CGroup () {}

public: // attributes

	scene_group_t		*m_pSG;

public: // operations

	virtual void Render (ERenderMode eMode)
	{
		CNode::Render( eMode == eWireframeSpecialGreen ? eWireframeSpecialRed : eMode );
	}

protected: // implementation

	virtual void DumpInTreeCtrl (HWND hTreeCtrl, HTREEITEM hParent = TVI_ROOT)
	{
		char		szString[1024];
		sprintf(szString, "%s (group)", m_strName.c_str());
		HTREEITEM	hItem = InsertItem(hTreeCtrl, 4, szString, hParent, (LPARAM)(CNode*)this);
		CNode::DumpInTreeCtrl(hTreeCtrl, hItem);
	}
};

//==========================================================================================

class CGeomObj : public CNode {

public: // construction/destruction

	CGeomObj (CNode *pParent, scene_geomobj_t *pGeomObj);
	virtual ~CGeomObj ()
	{
		if ( m_pInterpolatedMeshVerts )
			free(m_pInterpolatedMeshVerts);
		if ( m_pInterpolatedMeshNormals )
			free(m_pInterpolatedMeshNormals);
		if ( m_pRenderableVerts )
			free(m_pRenderableVerts);
	}

public: // attributes

	//-----------------------------------------
	struct SMesh {
		scene_mesh_t					*m_pSM;
		AnimPoint3_t					*m_pVerts; // * m_pSM->num_verts
		AnimFace_t						*m_pFaces; // * m_pSM->num_faces
		vector<int>						m_vFaceMaterials; // * m_pSM->num_faces
		AnimFaceNormal_t				*m_pNormals; // * m_pSM->num_faces
		AnimUVVert_t					*m_pTVerts; // * m_pSM->num_tverts
		AnimTVFace_t					*m_pTFaces; // * m_pSM->num_faces
	};
	struct SPhysVert {
		union {
			scene_rigidvert_t		*m_pSRV;
			scene_blendedvert_t		*m_pSBV;
		};
		vector<CNode *>			m_vBones;
		vector<float>			m_vWeights;
	};
	//-----------------------------------------

	scene_geomobj_t		*m_pSGO;
	matrix_t			m_mNodeTM;
	matrix_t			m_mFigureTM;

	struct {
		scene_keys_t		*m_pSK;
		union {
			TCB_POSkey_t	*m_pTCBpk;
			HYB_POSkey_t	*m_pHYBpk;
			LIN_POSkey_t	*m_pLINpk;
		};
		union {
			TCB_ROTkey_t	*m_pTCBrk;
			HYB_ROTkey_t	*m_pHYBrk;
			LIN_ROTkey_t	*m_pLINrk;
		};
		union {
			TCB_SCLkey_t	*m_pTCBsk;
			HYB_SCLkey_t	*m_pHYBsk;
			LIN_SCLkey_t	*m_pLINsk;
		};
	} Keys;

	struct {
		scene_samptm_t		*m_pSS;
		TMSample_t			*m_pTM;
	} Samples;

	vector<SMesh>			m_vMeshes;

	scene_physverts_t		*m_pPhysVerts;
	vector<SPhysVert>		m_vPhysVerts;

protected: // object state

	// NODETYPE_DEFORMATION

	BOOL					m_bInterpolableMesh;
	const BOOL				m_bInterpolateNormals;

	AnimPoint3_t			*m_pInterpolatedMeshVerts; // consistente se m_pDeformationMesh == NULL
	AnimFaceNormal_t		*m_pInterpolatedMeshNormals; // consistente se m_pDeformationMesh == NULL e se m_bInterpolateNormals != FALSE
	SMesh					*m_pDeformationMesh; // se != NULL  comunque valido, a prescindere da m_pInterpolatedMesh...

	// ALL TYPES

	AnimPoint3_t			*m_pRenderableVerts;

public: // operations

	virtual void EvalWorldState (TimeValue time);
	virtual void GenerateRenderableVerts ();
	virtual void Render (ERenderMode eMode);

protected: // implementation / helpers

	matrix_t		m_mTM;
	matrix_t		m_mFinalTM;

	virtual void PostProcess ();

	static void Interpolate (AnimPoint3_t *pF, AnimPoint3_t *pS, float fI, AnimPoint3_t *pO)
	{
		vector_t		vT;
		VectorSubtract(pS,pF,&vT);
		vector_t		vA;
		VectorMultByScalar(&vT,fI,&vA);
		VectorAdd(pF,&vA,pO);
	}

protected: // implementation

	virtual void DumpInTreeCtrl (HWND hTreeCtrl, HTREEITEM hParent = TVI_ROOT)
	{
		char		szString[1024];
		strcpy(szString, m_strName.c_str());
		switch (m_pSGO->node_type) {
		case NODETYPE_BIPEDBONE: strcat(szString, " (biped bone)"); break;
		case NODETYPE_RIGID: strcat(szString, " (rigid)"); break;
		case NODETYPE_DEFORMATION: strcat(szString, " (deformation)"); break;
		case NODETYPE_PHYSIQUE: strcat(szString, " (physique)"); break;
		}
		HTREEITEM	hItem = InsertItem(hTreeCtrl, 5, szString, hParent, (LPARAM)(CNode*)this);

		sprintf(szString, "Node Identifier: %08lX", m_pSGO->node_id);
		InsertItem(hTreeCtrl, 0, szString, hItem);

		sprintf(szString, "Inherit Pos: X=%s; Y=%s; Z=%s", m_pSGO->inherit_pos[0] ? "Yes" : "No", m_pSGO->inherit_pos[1] ? "Yes" : "No", m_pSGO->inherit_pos[2] ? "Yes" : "No");
		InsertItem(hTreeCtrl, 0, szString, hItem);
		sprintf(szString, "Inherit Rot: X=%s; Y=%s; Z=%s", m_pSGO->inherit_rot[0] ? "Yes" : "No", m_pSGO->inherit_rot[1] ? "Yes" : "No", m_pSGO->inherit_rot[2] ? "Yes" : "No");
		InsertItem(hTreeCtrl, 0, szString, hItem);
		sprintf(szString, "Inherit Scl: X=%s; Y=%s; Z=%s", m_pSGO->inherit_scl[0] ? "Yes" : "No", m_pSGO->inherit_scl[1] ? "Yes" : "No", m_pSGO->inherit_scl[2] ? "Yes" : "No");
		InsertItem(hTreeCtrl, 0, szString, hItem);

		if ( Keys.m_pSK != NULL ) {
			strcpy(szString, "Position Controller: ");
			if ( Keys.m_pSK->position_controller == POSITIONCONTROLLER_TCB ) strcat(szString, "TCB");
			else if ( Keys.m_pSK->position_controller == POSITIONCONTROLLER_HYBRID ) strcat(szString, "Hybrid");
			else if ( Keys.m_pSK->position_controller == POSITIONCONTROLLER_LIN ) strcat(szString, "Linear");
			InsertItem(hTreeCtrl, 0, szString, hItem);

			sprintf(szString, "Number Of Position Keys: %i key(s)", Keys.m_pSK->position_keys);
			InsertItem(hTreeCtrl, 0, szString, hItem);

			strcpy(szString, "Rotation Controller: ");
			if ( Keys.m_pSK->rotation_controller == ROTATIONCONTROLLER_TCB ) strcat(szString, "TCB");
			else if ( Keys.m_pSK->rotation_controller == ROTATIONCONTROLLER_HYBRID ) strcat(szString, "Hybrid");
			else if ( Keys.m_pSK->rotation_controller == ROTATIONCONTROLLER_LIN ) strcat(szString, "Linear");
			InsertItem(hTreeCtrl, 0, szString, hItem);

			sprintf(szString, "Number Of Rotation Keys: %i key(s)", Keys.m_pSK->rotation_keys);
			InsertItem(hTreeCtrl, 0, szString, hItem);

			strcpy(szString, "Scale Controller: ");
			if ( Keys.m_pSK->scale_controller == SCALECONTROLLER_TCB ) strcat(szString, "TCB");
			else if ( Keys.m_pSK->scale_controller == SCALECONTROLLER_HYBRID ) strcat(szString, "Hybrid");
			else if ( Keys.m_pSK->scale_controller == SCALECONTROLLER_LIN ) strcat(szString, "Linear");
			InsertItem(hTreeCtrl, 0, szString, hItem);

			sprintf(szString, "Number Of Scale Keys: %i key(s)", Keys.m_pSK->scale_keys);
			InsertItem(hTreeCtrl, 0, szString, hItem);
		}
		else if ( Samples.m_pSS != NULL ) {
			sprintf(szString, "Number Of Sampled Keys: %i key(s)", Samples.m_pSS->keys_num);
			InsertItem(hTreeCtrl, 0, szString, hItem);
		}

		if ( m_pSGO->node_type == NODETYPE_DEFORMATION && m_vMeshes.size()>1 ) {
			sprintf(szString, "Are Meshes Interpolable: %s", m_bInterpolableMesh ? "Yes" : "No");
			InsertItem(hTreeCtrl, 0, szString, hItem);
		}

		for ( int i=0; i<m_vMeshes.size(); i++ ) {
			sprintf(szString, "Mesh %i", i);
			HTREEITEM		hMesh = InsertItem(hTreeCtrl, 6, szString, hItem);
			sprintf(szString, "Time: %i ticks", (int)m_vMeshes[i].m_pSM->time);
			InsertItem(hTreeCtrl, 0, szString, hMesh);
			sprintf(szString, "Frame: %i", m_vMeshes[i].m_pSM->time / ((CSceneRoot*)m_pSceneRoot)->m_pGI->ticks_per_frame);
			InsertItem(hTreeCtrl, 0, szString, hMesh);
			sprintf(szString, "Number Of Vertices: %i vert(s)", m_vMeshes[i].m_pSM->num_verts);
			InsertItem(hTreeCtrl, 0, szString, hMesh);
			sprintf(szString, "Number Of Faces: %i tri(s)", m_vMeshes[i].m_pSM->num_faces);
			InsertItem(hTreeCtrl, 0, szString, hMesh);
			sprintf(szString, "Number Of Texture Vertices: %i tvert(s)", m_vMeshes[i].m_pSM->num_tverts);
			InsertItem(hTreeCtrl, 0, szString, hMesh);
			sprintf(szString, "Has Native Face And Vertex Normals: %s", m_vMeshes[i].m_pSM->has_normals ? "Yes" : "No");
			InsertItem(hTreeCtrl, 0, szString, hMesh);
			sprintf(szString, "Raw Mesh Size: %i byte(s)",
				m_vMeshes[i].m_pSM->num_verts * sizeof(AnimPoint3_t) +
				m_vMeshes[i].m_pSM->num_faces * sizeof(AnimFace_t) +
				(m_vMeshes[i].m_pSM->has_normals ? m_vMeshes[i].m_pSM->num_faces * sizeof(AnimFaceNormal_t) : 0) +
				(m_vMeshes[i].m_pSM->num_tverts ? (m_vMeshes[i].m_pSM->num_tverts * sizeof(AnimUVVert_t) + m_vMeshes[i].m_pSM->num_faces * sizeof(AnimTVFace_t)) : 0));
			InsertItem(hTreeCtrl, 0, szString, hMesh);
		}

		if ( m_vPhysVerts.size() ) {
			sprintf(szString, "Number Of Physique Vertices: %i vert(s)", m_vPhysVerts.size());
			InsertItem(hTreeCtrl, 0, szString, hItem);
		}

		CNode::DumpInTreeCtrl(hTreeCtrl, hItem);
	}
};

//==========================================================================================

#endif